@@ -11,19 +11,20 @@ import (
 	"k8s.io/apimachinery/pkg/types"
 	"k8s.io/client-go/util/retry"
 	client "sigs.k8s.io/controller-runtime/pkg/client"
-	logf "sigs.k8s.io/controller-runtime/pkg/runtime/log"
+	logf "sigs.k8s.io/controller-runtime/pkg/log"
 
 	nmstate "github.com/nmstate/kubernetes-nmstate/api/shared"
 	nmstatev1beta1 "github.com/nmstate/kubernetes-nmstate/api/v1beta1"
 	enactmentconditions "github.com/nmstate/kubernetes-nmstate/pkg/enactmentstatus/conditions"
+	"github.com/nmstate/kubernetes-nmstate/pkg/node"
 )
 
 var (
 	log = logf.Log.WithName("policyconditions")
 )
 
-func setPolicyProgressing(conditions *nmstate.ConditionList, message string) {
-	log.Info("setPolicyProgressing")
+func SetPolicyProgressing(conditions *nmstate.ConditionList, message string) {
+	log.Info("SetPolicyProgressing")
 	conditions.Set(
 		nmstate.NodeNetworkConfigurationPolicyConditionDegraded,
 		corev1.ConditionUnknown,
@@ -38,8 +39,8 @@ func setPolicyProgressing(conditions *nmstate.ConditionList, message string) {
 	)
 }
 
-func setPolicySuccess(conditions *nmstate.ConditionList, message string) {
-	log.Info("setPolicySuccess")
+func SetPolicySuccess(conditions *nmstate.ConditionList, message string) {
+	log.Info("SetPolicySuccess")
 	conditions.Set(
 		nmstate.NodeNetworkConfigurationPolicyConditionDegraded,
 		corev1.ConditionFalse,
@@ -54,8 +55,8 @@ func setPolicySuccess(conditions *nmstate.ConditionList, message string) {
 	)
 }
 
-func setPolicyNotMatching(conditions *nmstate.ConditionList, message string) {
-	log.Info("setPolicyNotMatching")
+func SetPolicyNotMatching(conditions *nmstate.ConditionList, message string) {
+	log.Info("SetPolicyNotMatching")
 	conditions.Set(
 		nmstate.NodeNetworkConfigurationPolicyConditionDegraded,
 		corev1.ConditionFalse,
@@ -70,8 +71,8 @@ func setPolicyNotMatching(conditions *nmstate.ConditionList, message string) {
 	)
 }
 
-func setPolicyFailedToConfigure(conditions *nmstate.ConditionList, message string) {
-	log.Info("setPolicyFailedToConfigure")
+func SetPolicyFailedToConfigure(conditions *nmstate.ConditionList, message string) {
+	log.Info("SetPolicyFailedToConfigure")
 	conditions.Set(
 		nmstate.NodeNetworkConfigurationPolicyConditionDegraded,
 		corev1.ConditionTrue,
@@ -86,85 +87,72 @@ func setPolicyFailedToConfigure(conditions *nmstate.ConditionList, message strin
 	)
 }
 
-func nodesRunningNmstate(cli client.Client) ([]corev1.Node, error) {
-	nodes := corev1.NodeList{}
-	err := cli.List(context.TODO(), &nodes)
-	if err != nil {
-		return []corev1.Node{}, errors.Wrap(err, "getting nodes failed")
-	}
+func Update(cli client.Client, apiReader client.Reader, policyKey types.NamespacedName) error {
+	logger := log.WithValues("policy", policyKey.Name)
 
-	pods := corev1.PodList{}
-	byComponent := client.MatchingLabels{"component": "kubernetes-nmstate-handler"}
-	err = cli.List(context.TODO(), &pods, byComponent)
+	err := update(cli, apiReader, cli, policyKey)
 	if err != nil {
-		return []corev1.Node{}, errors.Wrap(err, "getting pods failed")
-	}
-
-	filteredNodes := []corev1.Node{}
-	for _, node := range nodes.Items {
-		for _, pod := range pods.Items {
-			if node.Name == pod.Spec.NodeName {
-				filteredNodes = append(filteredNodes, node)
-				break
-			}
+		logger.Error(err, "failed to update policy status using cached client. Retrying with non-cached.")
+		err = update(cli, apiReader, apiReader, policyKey)
+		if err != nil {
+			logger.Error(err, "failed to update policy status using non-cached client.")
 		}
 	}
-	return filteredNodes, nil
+	return err
 }
 
-func Update(cli client.Client, policyKey types.NamespacedName) error {
+func update(apiWriter client.Client, apiReader client.Reader, policyReader client.Reader, policyKey types.NamespacedName) error {
 	logger := log.WithValues("policy", policyKey.Name)
 	// On conflict we need to re-retrieve enactments since the
 	// conflict can denote that the calculated policy conditions
 	// are now not accurate.
 	return retry.RetryOnConflict(retry.DefaultRetry, func() error {
 		policy := &nmstatev1beta1.NodeNetworkConfigurationPolicy{}
-		err := cli.Get(context.TODO(), policyKey, policy)
+		err := policyReader.Get(context.TODO(), policyKey, policy)
 		if err != nil {
 			return errors.Wrap(err, "getting policy failed")
 		}
 
 		enactments := nmstatev1beta1.NodeNetworkConfigurationEnactmentList{}
 		policyLabelFilter := client.MatchingLabels{nmstate.EnactmentPolicyLabel: policy.Name}
-		err = cli.List(context.TODO(), &enactments, policyLabelFilter)
+		err = apiReader.List(context.TODO(), &enactments, policyLabelFilter)
 		if err != nil {
 			return errors.Wrap(err, "getting enactments failed")
 		}
 
-		// Count only nodes that runs nmstate handler, could be that
-		// users don't want to run knmstate at master for example so
-		// they don't want to change net config there.
-		nmstateNodes, err := nodesRunningNmstate(cli)
+		// Count only nodes that runs nmstate handler and match the policy
+		// nodeSelector, could be that users don't want to run knmstate at control-plane for example
+		// so they don't want to change net config there.
+		nmstateMatchingNodes, err := node.NodesRunningNmstate(apiReader, policy.Spec.NodeSelector)
 		if err != nil {
 			return errors.Wrap(err, "getting nodes running kubernets-nmstate pods failed")
 		}
-		numberOfNmstateNodes := len(nmstateNodes)
+		numberOfNmstateMatchingNodes := len(nmstateMatchingNodes)
 
 		// Let's get conditions with true status count filtered by policy generation
-		enactmentsCount := enactmentconditions.Count(enactments, policy.Generation)
+		enactmentsCountByCondition := enactmentconditions.Count(enactments, policy.Generation)
 
-		numberOfFinishedEnactments := enactmentsCount.Available() + enactmentsCount.Failed() + enactmentsCount.NotMatching() + enactmentsCount.Aborted()
+		numberOfFinishedEnactments := enactmentsCountByCondition.Available() + enactmentsCountByCondition.Failed() + enactmentsCountByCondition.Aborted()
 
-		logger.Info(fmt.Sprintf("enactments count: %s", enactmentsCount))
-		if numberOfFinishedEnactments < numberOfNmstateNodes {
-			setPolicyProgressing(&policy.Status.Conditions, fmt.Sprintf("Policy is progressing %d/%d nodes finished", numberOfFinishedEnactments, numberOfNmstateNodes))
-		} else {
-			if enactmentsCount.Matching() == 0 {
-				message := "Policy does not match any node"
-				setPolicyNotMatching(&policy.Status.Conditions, message)
-			} else if enactmentsCount.Failed() > 0 || enactmentsCount.Aborted() > 0 {
-				message := fmt.Sprintf("%d/%d nodes failed to configure", enactmentsCount.Failed(), enactmentsCount.Matching())
-				if enactmentsCount.Aborted() > 0 {
-					message += fmt.Sprintf(", %d nodes aborted configuration", enactmentsCount.Aborted())
-				}
-				setPolicyFailedToConfigure(&policy.Status.Conditions, message)
-			} else {
-				message := fmt.Sprintf("%d/%d nodes successfully configured", enactmentsCount.Available(), enactmentsCount.Available())
-				setPolicySuccess(&policy.Status.Conditions, message)
+		logger.Info(fmt.Sprintf("numberOfNmstateMatchingNodes: %d, enactments count: %s", numberOfNmstateMatchingNodes, enactmentsCountByCondition))
+
+		if numberOfNmstateMatchingNodes == 0 {
+			message := "Policy does not match any node"
+			SetPolicyNotMatching(&policy.Status.Conditions, message)
+		} else if enactmentsCountByCondition.Failed() > 0 || enactmentsCountByCondition.Aborted() > 0 {
+			message := fmt.Sprintf("%d/%d nodes failed to configure", enactmentsCountByCondition.Failed(), numberOfNmstateMatchingNodes)
+			if enactmentsCountByCondition.Aborted() > 0 {
+				message += fmt.Sprintf(", %d nodes aborted configuration", enactmentsCountByCondition.Aborted())
 			}
+			SetPolicyFailedToConfigure(&policy.Status.Conditions, message)
+		} else if numberOfFinishedEnactments < numberOfNmstateMatchingNodes {
+			SetPolicyProgressing(&policy.Status.Conditions, fmt.Sprintf("Policy is progressing %d/%d nodes finished", numberOfFinishedEnactments, numberOfNmstateMatchingNodes))
+		} else {
+			message := fmt.Sprintf("%d/%d nodes successfully configured", enactmentsCountByCondition.Available(), enactmentsCountByCondition.Available())
+			SetPolicySuccess(&policy.Status.Conditions, message)
 		}
 
-		err = cli.Status().Update(context.TODO(), policy)
+		err = apiWriter.Status().Update(context.TODO(), policy)
 		if err != nil {
 			if apierrors.IsConflict(err) {
 				logger.Info("conflict updating policy conditions, retrying")
